from io import StringIO
from rosbags.rosbag2 import Reader
from pathlib import Path
from rosbags.serde import deserialize_cdr
import logging
import json
from .reader_parser import Parser
import numpy as np
import os

CHUNK_SIZE = 50 * 1024 * 1024

###############
##### PG ######
###############


class NumpyEncoder(json.JSONEncoder):
    """
    Overrides the default() method to handle NumPy arrays.

    Parameters:
    -----------
    obj : object
        Object to be serialized.

    Returns:
    --------
    list or object
        If obj is a NumPy array, returns a list containing the elements of the array.
        Otherwise, returns the result of the default JSONEncoder's default method.
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class BagReaderCustomMessages:

    """
    Custom reader for ROS bag files with support for specialized message types.

    This class is responsible for reading messages from ROS bag files and serializing
    them into JSON format. It handles a variety of message types, including custom ones,
    and manages various complexities like nested message structures and special values.

    Attributes:
    -----------
    log : Logger
        A logging object for capturing runtime information.
    parser : Parser
        A parser object used to convert deserialized ROS messages to dictionary format.
    unsupported_topics : list
        A list to keep track of unsupported topics encountered during processing.

    Methods:
    --------
    sanitize_json(message_json: dict) -> dict:
        Sanitizes the JSON object by replacing certain special values like 'NaN' and 'Infinity'.

    handle_tf_msgs(msg: dict) -> dict:
        Handles transformation messages, reshaping them for easier processing.

    read_messages(input_bag: str, simulation_run_id: int) -> StringIO:
        Reads messages from a ROS bag file, serializes them into JSON, and writes them to a buffer.

    Example Usage:
    --------------
    >>> reader = BagReaderCustomMessages('path/to/custom/messages')
    >>> buffer = reader.read_messages('path/to/bagfile.bag', 1)
    """

    def __init__(
        self,
        custom_message_path,
        log=None,
        debug=False,
        verbose=False,
    ):
        self.log = log
        self.debug = debug
        self.verbose = verbose

        self.log.debug(f"custom_message_path = {custom_message_path}")

        self.parser = Parser(custom_message_path, self.log)

        self.unsupported_topics = []

    @staticmethod
    def sanitize_json(message_json: dict) -> dict:
        if "NaN" in message_json:
            message_json = message_json.replace("NaN", "-1")
        if "Infinity" in message_json:
            message_json = message_json.replace("Infinity", "-1")
        if "<?xml version=" in message_json:  # robot description topic
            message_json = json.dumps(
                {"error": "robot description not supported yet ..."}
            )
        return message_json

    @staticmethod
    def handle_tf_msgs(msg: dict) -> dict:
        message_dict = {}
        for tf in msg["transforms"]:
            message_dict[tf.header.frame_id] = {
                "header": str(tf.header),
                "__msgtype__": tf.__msgtype__,
                "translation": [
                    tf.transform.translation.x,
                    tf.transform.translation.y,
                    tf.transform.translation.z,
                ],
                "rotation": [
                    tf.transform.rotation.x,
                    tf.transform.rotation.y,
                    tf.transform.rotation.z,
                    tf.transform.rotation.w,
                ],
            }
        return message_dict

    def read_messages(self, input_bag: str, simulation_run_id: int) -> StringIO:
        buffer = StringIO()
        size = 0
        rid = 0

        file_extension = input_bag.split(".")[1]
        self.log.debug(
            f"sqlite3 bag detected, {input_bag}"
        ) if file_extension != "mcap" else self.log.debug(
            f"mcap bag detected, {input_bag}"
        )
        input_bag = os.path.dirname(input_bag)

        with Reader(Path(input_bag)) as reader:
            connections = [
                x
                for x in reader.connections
                if x.topic not in ["/rosout", "/client_count", "/connected_clients"]
            ]
            for connection, timestamp, raw_data in reader.messages(
                connections=connections
            ):
                if (
                    connection.msgtype == "rcl_interfaces/msg/ParameterEvent"
                ):  # ignoring parameter event topic
                    continue

                try:
                    message_deserialized = deserialize_cdr(raw_data, connection.msgtype)
                    message_dict = self.parser.deserialized_message_to_dict(
                        message_deserialized
                    )
                except KeyError as e:  # unregister message type, skipping
                    if connection.topic not in self.unsupported_topics:
                        self.log.warning(
                            f"[KeyError] skipping topic: {connection.topic} because unsupported message type: {connection.msgtype}, if nested, unknown: {str(e)}"
                        )
                        self.unsupported_topics.append(connection.topic)
                    continue

                except AssertionError as e:
                    self.log.warning(
                        f"[AssertionError]: skipping topic: {connection.topic} because unsupported message type: {connection.msgtype}, if nested, unknown: {str(e)}"
                    )
                    continue

                if (
                    "Multi" in connection.msgtype
                ):  # message type is MultiArray, take only data and __msgtype__
                    message_dict = {
                        k: message_dict[k]
                        for k in message_dict.keys() & {"data", "__msgtype__"}
                    }

                if "sensor_msgs/msg/Image" in connection.msgtype:
                    self.log.warning("Image message detected ...")
                    continue

                if "tf2_msgs/msg/TFMessage" in connection.msgtype:
                    message_dict = self.handle_tf_msgs(message_dict)

                try:  # parse message
                    message_json = json.dumps(message_dict, cls=NumpyEncoder)
                except Exception:
                    try:  # nested message, try to flatten it
                        message_dict = self.parser.flatten_dict(message_dict)
                        message_dict = self.parser.nest_dict(message_dict)
                        message_json = json.dumps(message_dict, cls=NumpyEncoder)
                    except (
                        Exception
                    ):  # can't flatten the message, need further inspection
                        if connection.topic not in self.unsupported_topics:
                            self.log.warning(
                                f"ignoring topic: {connection.topic}, with type: {connection.msgtype}"
                            )
                            self.unsupported_topics.append(connection.topic)
                        continue

                message_json = self.sanitize_json(message_json)

                try:
                    row = chr(0x1E).join(
                        [
                            f"{simulation_run_id}",
                            f"{rid}",
                            f"{timestamp}",
                            f"{connection.topic}",
                            f"{connection.msgtype}",
                            message_json,
                        ]
                    )
                except Exception:  # not a valid json, need further inspection
                    self.log.error(f"message_json = {message_json}")
                    raise

                rid = rid + 1
                bytes_wrote = buffer.write(row + "\n")
                size = size + bytes_wrote
                # 50MB chunks max
                if size >= CHUNK_SIZE:
                    yield buffer
                    buffer, size = StringIO(), 0
        # yielding the remaining content.
        yield buffer



import numpy as np
import json


class CitrosDict(dict):
    """
    Citros dictionary

    CitrosDict is a dictionary-like object, that allows to print the content as a json-object.
    """

    def to_json(self):
        """
        Convert to json string.

        Returns
        -------
        out : str
            json_str

        Examples
        --------
        Make a CitrosDict object, convert it to json string and print it:

        >>> d = da.CitrosDict({'package': 'data_analysis', 'module': 'data_access', 'object': 'CitrosDict', 'style': 'json'})
        >>> print(d.to_json())
        {
          "package": "data_analysis",
          "module": "data_access",
          "object": "CitrosDict",
          "style": "json"
        }
        """
        json_str = json.dumps(self, indent=2, cls=_NpEncoder)
        return json_str

    def print(self):
        """
        Print content of the CitrosDict object in a 'json'-style.

        Examples
        --------
        Make a CitrosDict object and print it in json-style:

        >>> d = da.CitrosDict({'package': 'data_analysis', 'module': 'data_access', 'object': 'CitrosDict', 'style': 'json'})
        >>> d.print()
        {
         'package': 'data_analysis',
         'module': 'data_access',
         'object': 'CitrosDict',
         'style': 'json'
        }
        """
        print("{")
        self._print()
        print("}")

    def _print(self, indent=" "):
        """
        Recursively handles and print items of the CitrosDict object.

        Parameters
        ----------
        str
            Indent for the current row.
        """
        if isinstance(self, dict):
            N_len = len(self.keys())
            for i, (k, v) in enumerate(self.items()):
                if isinstance(v, dict):
                    if isinstance(k, str):
                        row_key = "'" + k + "'"
                    else:
                        row_key = str(k)
                    print(indent + row_key + ": {")
                    v._print(indent + "  ")
                    if i == N_len - 1:
                        print(indent + "}")
                    else:
                        print(indent + "},")
                else:
                    if isinstance(k, str):
                        row_key = "'" + k + "'"
                    else:
                        row_key = str(k)
                    if isinstance(v, str):
                        row_val = "'" + v + "'"
                    else:
                        row_val = str(v)
                    if i == N_len - 1:
                        print(indent + row_key + ": " + row_val)
                    else:
                        print(indent + row_key + ": " + row_val + ",")

    def _get_type_dict(self, type_dict, data_dict):
        """
        Recursively find types of the dict values of the `data_dict` and write them in `type_dict`.

        Dictionaries may be embedded in dictionaries.

        Parameters
        ----------
        type_dict : dict
            dict in which the result is written.
        data_dict : dict
            dict to find types of its values.
        """
        for k, v in data_dict.items():
            if isinstance(v, dict):
                type_dict[k] = CitrosDict()
                CitrosDict()._get_type_dict(type_dict[k], v)
            else:
                type_dict[k] = type(v).__name__


class _NpEncoder(json.JSONEncoder):
    """
    Handle numpy types.
    """

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(_NpEncoder, self).default(obj)



# to run: python3 -m pytest test_parse_xml.py

import pytest
import random
import os
import tempfile
from faker import Faker
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.dom import minidom
from citros import Citros

NUM_TEST_RUNS = 10


def generate_package_xml(
    package_name,
    version,
    maintainer,
    maintainer_email,
    description,
    license,
    build_type,
):
    """
    Generates a package.xml string.
    """
    package = Element("package")

    name_elem = SubElement(package, "name")
    name_elem.text = package_name

    version_elem = SubElement(package, "version")
    version_elem.text = version

    maintainer_elem = SubElement(package, "maintainer", {"email": maintainer_email})
    maintainer_elem.text = maintainer

    description_elem = SubElement(package, "description")
    description_elem.text = description

    license_elem = SubElement(package, "license")
    license_elem.text = license

    export_elem = SubElement(package, "export")
    build_type_elem = SubElement(export_elem, "build_type")
    build_type_elem.text = build_type

    raw_string = tostring(package, "utf-8")
    reparsed = minidom.parseString(raw_string)
    return reparsed.toprettyxml(indent="  ")


def generate_random_version():
    return f"{random.randint(1, 10)}.{random.randint(0, 9)}.{random.randint(0, 9)}"


def generate_test_data():
    fake = Faker()
    for _ in range(NUM_TEST_RUNS):
        yield (
            fake.word(),
            generate_random_version(),
            fake.name(),
            fake.email(),
            fake.sentence(),
            "Apache-2.0",
            random.choice(["ament_cmake", "ament_python"]),
        )


@pytest.mark.parametrize(
    "package_name, version, maintainer, maintainer_email, description, license, build_type",
    generate_test_data(),
)
def test_parse_xml(
    package_name,
    version,
    maintainer,
    maintainer_email,
    description,
    license,
    build_type,
):
    package_xml_content = generate_package_xml(
        package_name,
        version,
        maintainer,
        maintainer_email,
        description,
        license,
        build_type,
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        xml_path = os.path.join(tmpdir, "package.xml")
        with open(xml_path, "w") as f:
            f.write(package_xml_content)
            f.close()
            package_path = os.path.dirname(f.name)

            with Citros() as citros:
                # Test the parse_xml function
                result = citros.parser_ros2.parse_xml(package_path)

                # Verify the results
                assert result == {
                    "package_xml": f.name,
                    "package_name": package_name,
                    "version": version,
                    "maintainer": maintainer,
                    "maintainer_email": maintainer_email,
                    "description": description,
                    "license": license,
                    "nodes": [],
                    "build_type": build_type,
                }


def test_empty_file():
    with tempfile.TemporaryDirectory() as tmpdir:
        xml_path = os.path.join(tmpdir, "package.xml")
        with open(xml_path, "w") as f:
            f.close()  # Empty file

            with Citros() as citros:
                result = citros.parser_ros2.parse_xml(os.path.dirname(f.name))

        assert result == {}


def test_missing_fields():
    package_xml_content = """
    <?xml version="1.0"?>
    <?xml-model href="http://download.ros.org/schema/package_format2.xsd" schematypens="http://www.w3.org/2001/XMLSchema"?>
    <package format="2">
        <name>rclcpp_tutorials</name>
    </package>
    """.strip()

    with tempfile.TemporaryDirectory() as tmpdir:
        xml_path = os.path.join(tmpdir, "package.xml")
        with open(xml_path, "w") as f:
            f.write(package_xml_content)
            f.close()

            with Citros() as citros:
                result = citros.parser_ros2.parse_xml(os.path.dirname(f.name))

            assert result == {
                "package_xml": f.name,
                "package_name": "rclcpp_tutorials",
                "version": "",
                "maintainer": "",
                "maintainer_email": "",
                "description": "",
                "license": "",
                "nodes": [],
                "build_type": None,
            }


def test_invalid_xml():
    package_xml_content = """
    This is not a valid XML file.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        xml_path = os.path.join(tmpdir, "package.xml")
        with open(xml_path, "w") as f:
            f.write(package_xml_content)
            f.close()

            with Citros() as citros:
                result = citros.parser_ros2.parse_xml(os.path.dirname(f.name))

            assert result == {}


def test_gleb_xml():
    with Citros() as citros:
        result = citros.parser_ros2.parse_xml("input/gleb")

        assert result == {
            "package_xml": "input/gleb/package.xml",
            "package_name": "turtlebot3_fake_node",
            "version": "2.2.3",
            "maintainer": "Will Son",
            "maintainer_email": "willson@robotis.com",
            "description": "\n    Package for TurtleBot3 fake node. With this package, simple tests can be done without a robot.\n"
            + "    You can do simple tests using this package on rviz without real robots.\n  ",
            "license": "Apache 2.0",
            "nodes": [],
            "build_type": "ament_cmake",
        }



import os
import sys
import glob
import json
import nbformat
import importlib_resources
from PyPDF2 import PdfReader, PdfWriter
from pathlib import Path
from datetime import datetime
from nbconvert import HTMLExporter, PDFExporter
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError
from rich import inspect
from weasyprint import HTML
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from .citros import Citros

from .logger import get_logger, shutdown_log


class NoNotebookFoundException(Exception):
    def __init__(self, message="No Notebook found."):
        super().__init__(message)


# citros report -n best_report_ever -notebook notebooks/notebook_name.ipynb -b vova_batch_1/version

# TODO: check that vova_batch_1/version is loaded, if not, load it
# reports/report_name/
# reports/report_name/notebooks/
# reports/report_name/batch/name (soft link)
# reports/report_name/info.json (status of report)
# reports/report_name/output/ (render output of notebooks + pdfs )


class Report:
    """

    Raises:
        NoNotebookFoundException: _description_
        NoNotebookFoundException: _description_

    Returns:
        _type_: _description_
    """

    ###################
    ##### private #####
    ###################
    def _init_log(self, log=None, verbose=False, debug=False):
        self.log = log
        self.verbose = verbose
        self.debug = debug
        if self.log is None:
            Path.home().joinpath(".citros/logs").mkdir(parents=True, exist_ok=True)
            log_dir = Path.home().joinpath(".citros/logs")

            self.log = get_logger(
                __name__,
                log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
                log_file=str(log_dir / "citros.log"),
                verbose=self.verbose,
            )

    def __init__(
        self,
        name: str = None,
        message: str = None,
        citros: Citros = None,
        output=None,
        batch=None,
        notebooks=[],
        sign=False,
        key=None,
        index: int = -1,  # default to take the last version of a runs
        version=None,
        log=None,
        debug=False,
        verbose=False,
    ):
        self._init_log(log, verbose, debug)
        self.log.debug(f"{self.__class__.__name__}.__init__()")
        self.sign_report = sign

        self.citros = citros
        self.batch = batch
        if self.batch is None:
            self.log.error("batch is None")
            return

        self.version = version
        self.index = index

        self.output = output
        self.reports_dir = citros.root_citros / "reports" / name
        if output is not None:
            self.reports_dir = Path(output)  # / "reports" / name

        # get version
        if not version:  # no version specified
            if batch:  # create new report
                self.version = datetime.today().strftime("%Y%m%d%H%M%S")
            else:
                versions = sorted(glob.glob(f"{str(self.reports_dir)}/*"))
                # get version from index
                self.version = versions[self.index].split("/")[-1]

        self.folder = self.reports_dir / self.version

        # event = {
        #     "type": "ERROR",
        #     "message": "Report not found",
        # }
        self.state = {
            "notebooks": notebooks,
            "data": [
                {
                    "simulation": batch["simulation"],
                    "batch": batch["name"],
                    "version": batch.version,
                }
            ],
            "status": "START",
            "events": [],
            "name": name,
            "message": message,
            "progress": 0,
            "started_at": datetime.today().strftime("%Y-%m-%d %H:%M:%S"),
            "finished_at": None,
        }

        Path(self.folder).mkdir(parents=True, exist_ok=True)
        (Path(self.folder) / "output").mkdir(parents=True, exist_ok=True)
        # (Path(self.folder) / "notebooks").mkdir(parents=True, exist_ok=True)

        self._save()

    def __str__(self):
        # print_json(data=self.data)
        return json.dumps(self.state, indent=4)

    def __getitem__(self, key):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.state[key]

    def get(self, key, default=None):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.data.state(key, default)

    def __setitem__(self, key, newvalue):
        self.state[key] = newvalue
        self._save()

    def _save(self):
        self.log.debug(
            f"{self.__class__.__name__}._save()",
        )
        self.state["updated_at"]: datetime.today().strftime("%Y-%m-%d %H:%M:%S")

        with open(self.path(), "w") as file:
            json.dump(self.state, file, indent=4, sort_keys=True)

    ###################
    ##### public ######
    ###################
    def path(self):
        """return the full path to the current main file.

        default to ".citros/project.json"

        Returns:
            str: the full path to the current main file.
        """
        return self.folder / "info.json"

    def run(self):
        self.log.debug(f"{self.__class__.__name__}.run()")
        self.start()
        self.proccess(0)
        self.log.debug("Start executing notebooks")
        self.execute(self.state["notebooks"], self.folder / "output")

        notebooks = glob.glob(f"{self.folder / 'output'}/*")

        self.log.debug("Start rendering notebooks")
        self.render(notebooks, self.folder)
        if self.sign_report:
            self.log.debug("Start signing notebooks")
            self.sign()
            self.log.debug("Start validating notebooks")
            self.validate()
        self.end()

        return self.folder

    # report status.
    def start(self):
        self.log.debug(f"{self.__class__.__name__}.start()")
        self.state["status"] = "START"
        # TODO: copy notebooks to report folder
        # TODO: create soft link to batch folder

    def proccess(self, progress):
        self.state["progress"] = progress

    def end(self, status="END"):
        self.log.debug(f"{self.__class__.__name__}.end()")
        self.proccess(100)
        self["status"] = status
        self["finished_at"] = datetime.today().strftime("%Y-%m-%d %H:%M:%S")

    def event(self, event, message):
        self.state["events"].append({"type": event, "message": message})

    ###################
    ##### public ######
    ###################
    def execute(self, notebook_paths, output_folder, timout=600):
        """
        This function executes jupiter notebooks provided
        Args:
            notebook_paths (str): path to folder with notebooks
            output_folder (str): path where executed notebooks should be

        """
        self.log.debug(f"{self.__class__.__name__}.execute_notebooks()")
        # config = Config()
        # config.ExecutePreprocessor.kernel_name = "python3"

        for notebook_path in notebook_paths:
            notebook_name = notebook_path.split("/")[-1]
            try:
                with open(notebook_path, "r", encoding="utf-8") as nb_file:
                    try:
                        nb_node = nbformat.read(nb_file, as_version=4)
                    except nbformat.reader.NotJSONError:
                        self.log.debug(f"The file {notebook_path} is not valid JSON.")
                        continue
                    except nbformat.validator.ValidationError as e:
                        self.log.debug(
                            f"The file {notebook_path} is not a valid notebook; validation error: {e}"
                        )
                        continue
            except FileNotFoundError:
                self.log.error(f"The file {notebook_path} does not exist.")
                raise NoNotebookFoundException

            # import os
            os.environ["REPORT_NANE"] = "CITROS"
            os.environ["REPORT_VERSION"] = "REPORT_VERSION"
            os.environ["BATCH_SIMULATION"] = "BATCH_SIMULATION"
            os.environ["BATCH_NAME"] = "BATCH_NAME"
            os.environ["BATCH_VERSION"] = "BATCH_VERSION"

            os.environ["PG_HOST"] = "localhost"
            os.environ["PG_PORT"] = "5454"
            os.environ["PG_DATABASE"] = "citros"
            os.environ["PG_SCHEMA"] = "citros"
            os.environ["PG_USER"] = "citros"
            os.environ["PG_PASSWORD"] = "password"

            os.environ["CITROS_ROOT"] = self.reports_dir

            execute_preprocessor = ExecutePreprocessor(
                timeout=timout, kernel_name="python3"
            )
            try:
                execute_preprocessor.preprocess(
                    nb_node,
                    {"metadata": {"path": output_folder}},
                )
            except CellExecutionError as e:
                self.log.exception(e)
                # raise
            finally:
                with open(
                    output_folder / notebook_name, "wt", encoding="utf-8"
                ) as nb_file:
                    nbformat.write(nb_node, nb_file)

    def render(self, notebook_paths, output_folder, css_file_path=None):
        """
        This function renders executed notebooks to PDF file.

        Args:
            notebook_paths (str): path to folder with notebooks
            output_folder (str): path where notebooks should be rendered
            css_file_path (str, optional): path to css file, defaults to 'data/reports/templates/default_style.css'.
        """
        self.log.debug(
            f"{self.__class__.__name__}.render_notebooks_to_pdf({notebook_paths})"
        )

        html_exporter = HTMLExporter(theme="light")
        # from jinja2 import DictLoader

        # with open(
        #     importlib_resources.files(f"data.reports").joinpath(
        #         "templates/index.html.j2"
        #     ),
        #     "r",
        # ) as template_file:
        #     html_template = template_file.read()

        # dl = DictLoader({"citros": html_template})

        # html_exporter = HTMLExporter(extra_loaders=[dl], template_file="citros")
        returns = []
        for notebook_path in notebook_paths:
            output_pdf_path = os.path.join(
                output_folder, os.path.basename(notebook_path).replace(".ipynb", ".pdf")
            )
            try:
                with open(notebook_path) as nb_file:
                    nb_node = nbformat.read(nb_file, as_version=4)
            except FileNotFoundError:
                self.log.error(f"The file {notebook_path} does not exist.")
                raise NoNotebookFoundException

            (body, _) = html_exporter.from_notebook_node(nb_node)

            output_html_path = os.path.join(
                output_folder,
                os.path.basename(notebook_path).replace(".ipynb", ".html"),
            )

            with open(output_html_path, "w") as html_file:
                html_file.write(body)
            # print(final_html)

            HTML(string=body).write_pdf(output_pdf_path)

            returns.append(output_pdf_path)
        return returns

    def sign(self, pdf_path, private_key_path, output_folder):
        """
        Signs PDF with private key

        Args:
            pdf_path (str): path to PDF file that needs to be signed
            private_key_path (str): path to private key
            output_folder (str): path to folder where signed pdf should be saved
        """
        raise NotImplementedError
        self.log.debug(f"{self.__class__.__name__}.sign_pdf_with_key()")
        with open(private_key_path, "rb") as key_file:
            private_key = serialization.load_pem_private_key(
                key_file.read(), password=None, backend=default_backend()
            )

        reader = PdfReader(pdf_path)
        content = b"".join(
            [page.extract_text().encode("utf-8") for page in reader.pages]
        )

        signature = private_key.sign(
            content,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256(),
        )

        output_pdf_path = os.path.join(output_folder, os.path.basename(pdf_path))
        writer = PdfWriter()
        for page in reader.pages:
            writer.add_page(page)
        writer.add_metadata({"/CitrosSign": signature})

        with open(output_pdf_path, "wb") as output_file:
            writer.write(output_file)

    def validate(self, pdf_path, public_key_path):
        """
        Checks if signed PDF was altered or not using public key

        Args:
            pdf_path (str): path to PDF files for check
            public_key_path (str): path to public key

        Returns:
            bool: Result of check
        """
        raise NotImplementedError
        self.log.debug(f"{self.__class__.__name__}.verify()")
        with open(public_key_path, "rb") as key_file:
            public_key = serialization.load_pem_public_key(
                key_file.read(), backend=default_backend()
            )

        reader = PdfReader(pdf_path)
        content = b"".join(
            [page.extract_text().encode("utf-8") for page in reader.pages]
        )
        signature = reader.metadata.get("/CitrosSign", None)

        if signature is None:
            return False

        try:
            public_key.verify(
                signature,
                content,
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH,
                ),
                hashes.SHA256(),
            )
            return True
        except Exception as e:
            self.log.exception(f"Verification failed: {e}")
            return False



import os

def get_version():
    """
    Return version of the citros_data_analysis package.

    Examples
    --------
    >>> from citros.citros_data_analysis import data_access as da
    >>> print(da.get_version())
    v0.1.1
    """
    rel_path = "../__init__.py"
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        for line in fp.read().splitlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        raise RuntimeError("Unable to find version string.")



# to run: python3 -m pytest test_parse_makefile.py

import os
import tempfile
from faker import Faker
import pytest
from citros import Citros

NUM_TEST_RUNS = 10

faker = Faker()


def generate_cmake_content(executables):
    content = "\n".join([f"add_executable({exe} src/{exe}.cpp)" for exe in executables])
    install_targets = " ".join(executables)
    content += (
        f"\n\ninstall(TARGETS {install_targets} DESTINATION lib/${{PROJECT_NAME}})\n"
    )
    return content


def cmake_data():
    for _ in range(NUM_TEST_RUNS):
        executables = [
            faker.pystr_format() for _ in range(faker.random_int(min=1, max=5))
        ]
        yield executables, generate_cmake_content(executables)


@pytest.mark.parametrize("executables,cmake_content", cmake_data())
def test_parse_makefile(executables, cmake_content):
    with tempfile.TemporaryDirectory() as tmpdir:
        cmake_path = os.path.join(tmpdir, "CMakeLists.txt")
        with open(cmake_path, "w") as f:
            f.write(cmake_content)

        with Citros(debug=True) as citros:
            result = citros.parser_ros2.parse_makefile(tmpdir)

        assert result["cmake"] == cmake_path
        nodes = result["nodes"]
        assert len(nodes) == len(executables)
        for node in nodes:
            assert "name" in node
            assert node["name"] in executables
            assert "entry_point" in node
            assert node["entry_point"] == ""
            assert "path" in node
            assert node["path"] == ""
            assert "parameters" in node
            assert node["parameters"] == []


def test_parse_makefile_no_install():
    with tempfile.TemporaryDirectory() as tmpdir:
        cmake_path = os.path.join(tmpdir, "CMakeLists.txt")
        with open(cmake_path, "w") as f:
            f.write("add_executable(foo src/foo.cpp)")

        with Citros(debug=True) as citros:
            with pytest.raises(ValueError) as e:
                citros.parser_ros2.parse_makefile(tmpdir)
            assert (
                str(e.value)
                == f"{cmake_path} is not formatted correctly: `add_executable` with no 'install' command."
            )


def test_parse_makefile_no_targets():
    with tempfile.TemporaryDirectory() as tmpdir:
        cmake_path = os.path.join(tmpdir, "CMakeLists.txt")
        with open(cmake_path, "w") as f:
            f.write("install(TARGETS DESTINATION lib/${PROJECT_NAME})")

        with Citros(debug=True) as citros:
            result = citros.parser_ros2.parse_makefile(tmpdir)

        # no targets (nodes) is OK - e.g. a cmake file that just
        # installs some common cmake functionality for other cmake files.
        assert result["cmake"] == cmake_path
        assert result["nodes"] == []


def test_parse_makefile_multiline_install():
    executables = [faker.pystr_format() for _ in range(faker.random_int(min=1, max=5))]

    with tempfile.TemporaryDirectory() as tmpdir:
        cmake_path = os.path.join(tmpdir, "CMakeLists.txt")
        with open(cmake_path, "w") as f:
            f.write("install(TARGETS \n")
            for exe in executables:
                f.write(f"{exe} \n")
            f.write("DESTINATION lib/${PROJECT_NAME})")

        with Citros(debug=True) as citros:
            result = citros.parser_ros2.parse_makefile(tmpdir)

        assert result["cmake"] == cmake_path
        nodes = result["nodes"]
        assert len(nodes) == len(executables)
        for node in nodes:
            assert "name" in node
            assert node["name"] in executables
            assert "entry_point" in node
            assert node["entry_point"] == ""
            assert "path" in node
            assert node["path"] == ""
            assert "parameters" in node
            assert node["parameters"] == []


def test_gleb_makefile():
    with Citros(debug=True) as citros:
        result = citros.parser_ros2.parse_makefile("input/gleb")

        assert result["cmake"] == "input/gleb/CMakeLists.txt"
        nodes = result["nodes"]
        assert len(nodes) == 1

        node = nodes[0]
        assert "name" in node
        assert node["name"] == "turtlebot3_fake_node"
        assert "entry_point" in node
        assert node["entry_point"] == ""
        assert "path" in node
        assert node["path"] == ""
        assert "parameters" in node
        assert node["parameters"] == []


def test_micro_ros_agent_makefile():
    with Citros(debug=True) as citros:
        result = citros.parser_ros2.parse_makefile("input/micro_ros_agent")

        assert result["cmake"] == "input/micro_ros_agent/CMakeLists.txt"
        nodes = result["nodes"]
        assert len(nodes) == 1

        node = nodes[0]
        assert "name" in node
        assert node["name"] == "micro_ros_agent"
        assert "entry_point" in node
        assert node["entry_point"] == ""
        assert "path" in node
        assert node["path"] == ""
        assert "parameters" in node
        assert node["parameters"] == []


def test_px4_makefile():
    with Citros(debug=True) as citros:
        result = citros.parser_ros2.parse_makefile("input/px4_msgs")

        assert result["cmake"] == "input/px4_msgs/CMakeLists.txt"
        nodes = result["nodes"]
        assert len(nodes) == 0


def test_multiple_assignments_makefile():
    with Citros(debug=True) as citros:
        result = citros.parser_ros2.parse_makefile("input/moveit2_planning")

        assert result["cmake"] == "input/moveit2_planning/CMakeLists.txt"

        nodes = result["nodes"]
        assert len(nodes) == 14


def test_multiple_installs_makefile():
    with Citros(debug=True) as citros:
        result = citros.parser_ros2.parse_makefile("input/moveit2_setup")

        assert result["cmake"] == "input/moveit2_setup/CMakeLists.txt"

        nodes = result["nodes"]
        assert len(nodes) == 1

        assert nodes[0]["name"] == "moveit_setup_simulation"



__title__ = "citros"
__description__ = "citros"
__url__ = "https://citros.io"
__version__ = "0.2.14"
__author__ = "lulav"
__author_email__ = "contact@lulav.space"
__license__ = "LICENSE"
__copyright__ = "Copyright 2024 Lulav Space"



from os import walk
from pathlib import Path
from rosbags.typesys import get_types_from_msg, register_types


class Parser:
    def __init__(self, custom_message_path="", log=None):
        self.log = log

        self.nested_dict_type = "rosbags.usertypes"
        if len(custom_message_path) > 0:
            for msg in custom_message_path:
                self.register_custom_messages(msg)

    def guess_msgtype(self, path: Path) -> str:
        """Guess message type name from path."""
        name = path.relative_to(path.parents[2]).with_suffix("")
        if "msg" not in name.parts:
            name = name.parent / "msg" / name.name
        return str(name)

    def register_custom_messages(self, message_folder_path):
        add_types = {}

        msgs = [next(walk(message_folder_path), (None, None, []))[2]][0]

        for pathstr in msgs:
            msgpath = Path(message_folder_path + pathstr)
            msgdef = msgpath.read_text(encoding="utf-8")
            add_types.update(get_types_from_msg(msgdef, self.guess_msgtype(msgpath)))
            self.log.debug(f"adding {msgpath} as custom message")

        register_types(add_types)

    def flatten_dict(
        self, dictionary: dict, parent_key: str = "", separator: str = "."
    ) -> dict:
        items = []
        for key, value in dictionary.items():
            new_key = parent_key + separator + key if parent_key else key
            try:
                if "/msg/" in value.__msgtype__:
                    value = self.deserialized_message_to_dict(value)
                    items.extend(
                        self.flatten_dict(value, new_key, separator=separator).items()
                    )
            except:
                items.append((new_key, value))
        return dict(items)

    def nest_dict(self, flat_dict):
        nested_dict = {}

        for key, value in flat_dict.items():
            keys = key.split(".")
            d = nested_dict
            for k in keys[:-1]:
                d = d.setdefault(k, {})
            d[keys[-1]] = value

        return nested_dict

    def deserialized_message_to_dict(self, dictionary: dict) -> dict:
        items = {}
        if type(dictionary) != "dict":
            dictionary = dictionary.__dict__
        for key, value in dictionary.items():
            value_type = str(type(value))
            if self.nested_dict_type in value_type:
                items[key] = self.deserialized_message_to_dict(value)
            else:
                items[key] = value
        return dict(items)
